#![no_std]
#![no_main]
#![feature(type_alias_impl_trait)]
#![feature(never_type)]

#[cfg(not(feature = "serial"))]
mod mouse_hid;

#[cfg(feature = "serial")]
mod usb_serial;

mod mouse_sensor;

use embassy_executor::Spawner;
use embassy_futures::join::join;
use embassy_rp::{
    peripherals::USB,
    spi::{CsPin, Instance},
};

#[cfg(not(feature = "serial"))]
use usbd_hid::descriptor::MouseReport;

#[cfg(not(feature = "serial"))]
use crate::{mouse_hid::mouse_config::MouseConfig, mouse_hid::MouseHID};

#[cfg(feature = "serial")]
use crate::usb_serial::Serial;

use crate::mouse_sensor::{MouseSensor, MouseSensorPins};

use {defmt_rtt as _, panic_probe as _};

#[embassy_executor::main]
async fn main(_spawner: Spawner) {
    let p = embassy_rp::init(Default::default());

    let mouse_sensor = MouseSensor::new(
        p.PIN_25,
        MouseSensorPins::new(p.SPI0, p.PIN_2, p.PIN_3, p.PIN_4, p.PIN_5),
        |_| (),
    )
    .await;

    #[cfg(not(feature = "serial"))]
    mouse_hid(p.USB, mouse_sensor).await;

    #[cfg(feature = "serial")]
    serial_hid(p.USB, mouse_sensor).await;
}

#[cfg(not(feature = "serial"))]
async fn mouse_hid<'d, L, T, CSP>(usb: USB, mut mouse_sensor: MouseSensor<'d, L, T, CSP>)
where
    L: FnMut(&str) + 'd,
    T: Instance,
    CSP: CsPin<T> + 'd,
{
    let mut mouse_hid = MouseHID::new(usb, MouseConfig::new());

    let usb_future = MouseHID::run_usb().await;

    // Do stuff with the class!
    let hid_future = async {
        loop {
            if let Some((x, y)) = mouse_sensor.read_movement().await {
                let report = MouseReport {
                    buttons: 0,
                    x,
                    y,
                    wheel: 0,
                    pan: 0,
                };

                #[cfg(not(feature = "serial"))]
                mouse_hid.mouse_report(report).await;
            }
        }
    };

    join(hid_future, usb_future).await;
}

#[cfg(feature = "serial")]
async fn serial_hid<'d, L, T, CSP>(usb: USB, mut mouse_sensor: MouseSensor<'d, L, T, CSP>)
where
    L: FnMut(&str) + 'd,
    T: Instance,
    CSP: CsPin<T> + 'd,
{
    let mut serial = Serial::new(usb);

    let usb_future = Serial::run_usb().await;

    // Do stuff with the class!
    let hid_future = async {
        loop {
            if let Some((x, y)) = mouse_sensor.read_movement().await {
                // let report = MouseReport {
                //     buttons: 0,
                //     x,
                //     y,
                //     wheel: 0,
                //     pan: 0,
                // };

                serial.send_msg("test").await;
            }
        }
    };

    join(hid_future, usb_future).await;
}
